import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.Border;
import java.util.*;
import java.io.*;
import java.net.*;

/**
 * Implementation of the Graffiti Client. 
 * Communicates with a GCP 1.0 Server
 */
public class GraffitiClient implements ActionListener {
	
	// UI components
	private JFrame frame;
	private JTextField text;
	private GraffitiDrawingArea gda; // Where images actually appear.
	
	// IO Stuff
	private StringTokenizer tok;
	private BufferedReader in;
		
	// socket stuff
	private Socket gcpSocket;
	private PrintWriter out;
	private String fromServer;
	private String fromUser;	
	private String IP;
	
	// constants for protocol messages 
	// B - Message from Server or Client
	// C - Message from Client
	// S - Message from Server
	public static final String C_PUTIMAGE = "PUT IMAGE";
	public static final String C_GETNEXT = "GET NEXT";
	public static final String C_GETPREV = "GET PREVIOUS";
	public static final String C_BYE = "BYE";
	public static final String B_IMAGESTART = "IMAGESTART";
	public static final String B_LINE = "LINE";
	public static final String B_IMAGEEND = "IMAGEEND";
	public static final String S_ERROR = "ERROR";
	public static final String S_BYE = "Bye.";
	public static final String S_BOL = "BOL";
	public static final String S_EOL = "EOL";
	public static final String S_OK = "OK";
	
	public GraffitiClient() { buildGUI(); }

	private void setupButton(JPanel p, String label, String command, String tip) {
        JButton b = new JButton(label);
        b.setActionCommand(command);
        b.setToolTipText(tip);
        b.addActionListener(this);
		p.add(b); // add button to this panel
	}

	private void buildGUI() {
		// Setup the button panel
		JPanel p = new JPanel();
		JPanel textPanel = new JPanel();
		setupButton(p,"Clear", "clear", "clear displayed image");
    	setupButton(p,"Next", "next", "get next image from Graffiti server");
    	setupButton(p,"Previous", "previous", "get previous image from Graffiti server");
    	setupButton(p,"Submit", "submit", "submit image to Graffiti server");

		// And now the drawing panel
		gda = new GraffitiDrawingArea();

		// Now the frame containing all panels
		frame = new JFrame("Graffiti Client");
		// Respond to close events
		
		frame.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent e) {
				
				// end converstation
				out.println(C_BYE);
				
				// close sockets and I/O 
				try {
					closeIO();
				} catch (IOException except) {
					System.err.println("Unable to close connection");
				}
				
				System.exit(0);
			}
		});
		
		frame.getContentPane().setLayout(new BorderLayout());
		Border empty;
		empty = BorderFactory.createEmptyBorder();
		
		// text field to display status of requests
		text = new JTextField(20);
		
		// setup textfield properties
		text.setEditable(false);
		text.setBorder(empty);
		text.setHorizontalAlignment(JTextField.CENTER);
		
		textPanel.add(text);
		
		frame.getContentPane().add(textPanel, BorderLayout.NORTH);	 // Add the text panel to the frame
		frame.getContentPane().add(gda, BorderLayout.CENTER); // Add the panel to the frame
		frame.getContentPane().add(p, BorderLayout.SOUTH); // Add the panel to the frame
		frame.pack();
		frame.setVisible(true);
	}
		
	/**
	 * Processes a button action chosen by the user.
	 * Performs according to GCP protocol specification.
	 * 
	 * @param	e	The button the user clicked.	
	 */
	public void actionPerformed(ActionEvent e) {
		//System.out.println("Action Performed: " +e.getActionCommand());
		
		if(e.getActionCommand().equals("clear")) {
			// clear the text field and image area
			clearPalette();
		}
		else if (e.getActionCommand().equals("submit")) {
			Enumeration enum = gda.image.elements();
			// Send header
			out.println(C_PUTIMAGE);
			out.println(B_IMAGESTART + " " + gda.getSize().width + 
						" " + gda.getSize().height); // send header
			
			// send lines from current image
			while (enum.hasMoreElements()) 
			{
				GraffitiDrawable gd = (GraffitiDrawable)enum.nextElement();
				out.println (B_LINE + " " + gd.getX1() + " " + gd.getY1() + 
							 " " + gd.getX2() + " " + gd.getY2());
			}
			// signal end of image
			out.println(B_IMAGEEND);			
			
			try {
				// check server response, if it is OK, everything
				// is ok, otherwise output the ERROR message
				
				fromServer = in.readLine();
					
				if (fromServer != null)
				{
					if (fromServer.startsWith(S_ERROR))					
					{
						displayErrorMsg(fromServer.substring(6, fromServer.length()));
						handleLostConnection();
					}
					else if (fromServer.equals(S_OK) == false)
					{
						displayErrorMsg("Error: Image Not Added");						
						handleLostConnection();
					}
				}					
			} catch (IOException excp) {
				displayErrorMsg("Error: I/O");
				handleLostConnection();
			}				
		}
		else if (e.getActionCommand().equals("previous")) {
			out.println(C_GETPREV);
			
			// process returned data
			readImageFromServer();			
		}
		else if (e.getActionCommand().equals("next")) {
			out.println(C_GETNEXT);
			
			// process returned data
			readImageFromServer();
		}		
	}
	
	/**
	 * Reads a single image from the server and
	 * then calls a method from the GraffitiDrawable
	 * instance to display the image
	 */
	public void readImageFromServer()
	{
		Vector newImage = new Vector();			
		int width = 0, height = 0, count = 0;

		String[] response = new String[6];
			
		clearStatusField();
						
		// process returned data
		try {
			fromServer = in.readLine();
			tok = new StringTokenizer(fromServer);
		} catch (IOException excp)	{
			displayErrorMsg("Error: I/O");
			handleLostConnection();
		}
			
		// first get the IMAGESTART header
		while (tok.hasMoreTokens())	
		{
			response[count] = tok.nextToken();
			count++;
		}			
			
		// if we get the correct header, then start processing lines
		if (response[0].equals(B_IMAGESTART))
		{
			width = Integer.parseInt(response[1]);
			height = Integer.parseInt(response[2]);
			
			// the below processes lines, we must catch IOExceptions
			// since we cannot throw them in this method.
			try {
				while ((fromServer = in.readLine()) != null)
				{
					count = 0;
					tok = new StringTokenizer(fromServer);
					
					while (tok.hasMoreTokens())	
					{
						response[count] = tok.nextToken();
						count++;
					}			
					
					// check what type of response we got from the 
					// server.  Should be a LINE x1 y1 x2 y2 or an
					// IMAGEEND
					if (response[0].equals(B_LINE))
					{
						// get coordinates for line
						int x1 = Integer.parseInt(response[1]);
						int y1 = Integer.parseInt(response[2]);
						int x2 = Integer.parseInt(response[3]);
						int y2 = Integer.parseInt(response[4]);

						// add the line to the new image
						newImage.add(new GraffitiLine(x1, y1, x2, y2));
					}
					else if (response[0].equals(B_IMAGEEND))
					{
						// load the image into the applet 
						// and exit the loop
						gda.loadImage(newImage);
						break;
					}
					else
					{
						displayErrorMsg("Error: Invalid Line");
						handleLostConnection();
					}
				}
							
			} catch (IOException excp) {
				displayErrorMsg("Error: I/O");
				handleLostConnection();
			}
					
		}
		else if (response[0].equals(S_BOL))
		{
			// if we get BOL, clear the applet and
			// output a msg
			gda.clear();

			setStatusField("No Previous Image");
		}			
		else if (response[0].equals(S_EOL))
		{
			// if we get EOL, clear the applet and
			// output a msg
			gda.clear();
			
			setStatusField("No More Images");
		}			
		else 
		{
			// A server side error occurred, we 
			// have lost the connection.
			displayErrorMsg("Error: Invalid Server Response");
			handleLostConnection();
		}
	}
	
	/**
	 * Displays the given error message on the console,
	 * in the status field, and in a popup box.
	 * 
	 * @param	msg	String containing error message
	 */
	public void displayErrorMsg(String msg)
	{
		setStatusField(msg);
		System.err.println(msg);
		JOptionPane.showMessageDialog(frame, msg, "Connection Lost", JOptionPane.INFORMATION_MESSAGE);		
	}
	
	/**
	 * Clears the text in the status field
	 */
    public void clearStatusField()
	{
		text.setText("");
	}
	
	/**
	 * Sets the text in the status field
	 * to the passed in string.
	 * 
	 * @param	txt	String to place in status field
	 */
    public void setStatusField(String txt)
	{
		text.setText(txt);
	}
	
	/**
	 * Resets drawing area and status field to
	 * it's initial state (empty).
	 */
	public void clearPalette()
	{
		clearStatusField();
		gda.clear();
	}
	
	/**
	 * Handles the connection to the Graffiti Server.
	 * Calls handleLostConnection() when it cannot connect to the server.
	 * 
	 * @see #handleLostConnection
	 */
	private void connectToServer()
	{
		// create socket
        gcpSocket = null;
        out = null;
        in = null;

        try {
            gcpSocket = new Socket(IP, 7777);
            out = new PrintWriter(gcpSocket.getOutputStream(), true);
            in = new BufferedReader(new InputStreamReader(gcpSocket.getInputStream()));
			
			fromServer = in.readLine();
			System.out.println("Server: " + fromServer);
	    } catch (UnknownHostException e) {
			displayErrorMsg("Could not find host: " + IP);
            handleLostConnection();
        } catch (IOException e) {
            displayErrorMsg("Couldn't get I/O for the connection to: " + IP);
			handleLostConnection();
        }
	}
	
	/**
	 * Gives the user the option to reconnect to the server
	 * by popping up a YES/NO option pane.  If the user chooses
	 * YES, connectToServer() is called, otherwise the system
	 * exists.
	 * 
	 * @see #connectToServer
	 */
	public void handleLostConnection()
	{			
		int choice;
		choice = JOptionPane.showConfirmDialog(frame, 
				"Would You Like to Reconnect?", "Question",
				JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE);
		
		if (choice == JOptionPane.YES_OPTION)
		{
			connectToServer();
			clearPalette();
		}
		else
			System.exit(0);
	}
	
	/**
	 * Closes I/O Connections
	 * 
	 * @throws	IOException	On trying to close sockets
	 */
	private void closeIO() throws IOException {
        out.close();
		in.close();
		gcpSocket.close();
	}

	/**
	 * Instantiates a GraffitiClient.  Then popup an Input dialog
	 * asking the user for a server IP to connect to.  If the user
	 * clicks OK, call connectToServer(), otherwise, quit.
	 * 
	 * @param args	Command line arguments
	 * @throws IOException
	 */
	public static void main(String[] args) throws IOException {
		GraffitiClient gc = new GraffitiClient();
		
		// get IP for this session
		gc.IP = JOptionPane.showInputDialog("Please enter a host to connect to:");
					
		if (gc.IP != null)
			gc.connectToServer();
		else
			System.exit(0);	
	}
}

class GraffitiDrawingArea extends JPanel {
	public Vector image = new Vector(); // An image consisting of a collection of
		// GraffitiDrawable instances

	private int xStart, yStart; // Track the start of a line segment

	public GraffitiDrawingArea(){ buildGUI(); }

	private void buildGUI(){
		setPreferredSize(new Dimension(400,400));
		setMaximumSize(new Dimension(400,400));
		setMinimumSize(new Dimension(400,400));
		
		addMouseListener(new MouseAdapter() {
			public void mousePressed(MouseEvent e) {
				// A click inside the drawing area indicates the start of
				// a line. Record it.
				xStart = e.getX(); yStart = e.getY();
			}
		});

		addMouseMotionListener(new MouseMotionAdapter() {
			public void mouseDragged(MouseEvent e){
				// User dragged to the end of a line. Record the line
				// in the image and schedule a paint of the image.
				// The repaint will include the new (additional) line.
				int x = e.getX(); int y = e.getY();
				image.add(new GraffitiLine(xStart, yStart, x,y));
				xStart=x; yStart=y;
				repaint();
			}
		});
	}
	
	/**
	 * Returns the maximum size of the drawable area.
	 * 
	 * @return	Dimension	Dimensions of drawing area
	 */
	public Dimension getSize()
	{
		return getMaximumSize();
	}
	
	/**
	 * Clears the current image, and repaints the window.
	 */
	public void clear() { 
		image = new Vector();
		repaint(); 
	}

	/**
	 * Displays the passed in image in the drawable area.
	 * 
	 * @param v	The Vector representation of an image.
	 */
	public void loadImage(Vector v) {
		image = v;
		repaint();
	}

	public void paint(Graphics g) {
		// This method eventually gets called. It causes the drawing area to be repainted.
		// For our purposes, a repaint involves drawing all the lines in the image.
		super.paint(g);
		Graphics2D g2d =(Graphics2D)g;
		Enumeration e = image.elements();
		while(e.hasMoreElements()){
			GraffitiDrawable gd=(GraffitiDrawable)e.nextElement();
			gd.draw(g2d);
		}
	}
}

interface GraffitiDrawable {
	
	public void draw(Graphics2D g);
	
	// methods to access the coords
	public int getX1();
	public int getY1();
	public int getX2();
	public int getY2();
}

class GraffitiLine implements GraffitiDrawable {
	private int x1, y1, x2, y2;
	
	GraffitiLine(int x1, int y1, int x2, int y2){
		this.x1=x1;
		this.y1=y1;
		this.x2=x2;
		this.y2=y2;
	}
	
	public void draw(Graphics2D g){
		g.drawLine(x1, y1, x2, y2);
	}
	
	/**
	 * Gets Line Coordinate x1
	 * 
	 * @returns x1	x coordinate of start position of line.
	 */
	public int getX1() { return x1; };
	
	/**
	 * Gets Line Coordinate y1
	 * 
	 * @returns y1	y coordinate of start position of line.
	 */
	public int getY1() { return y1; };
	/**
	 * Gets Line Coordinate x2
	 * 
	 * @returns x2	x coordinate of end position of line.
	 */
	public int getX2() { return x2; };

	/**
	 * Gets Line Coordinate y2
	 * 
	 * @returns y2	y coordinate of end position of line.
	 */
	public int getY2() { return y2; };
}